#include <linux/init.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/config.h>
#include <linux/slab.h> /* kmalloc() */
#include <linux/fs.h> /* everything... */
#include <linux/errno.h> /* error codes */
#include <linux/types.h> /* size_t */
#include <linux/buffer_head.h>
#include <linux/bio.h>
#include <linux/blkdev.h>
#include <linux/proc_fs.h>
#include <linux/time.h>
#include <linux/genhd.h>
#include <linux/fcntl.h> /* O_ACCMODE */
#include <asm/system.h> /* cli(), *_flags */
#include <asm/uaccess.h> /* copy_from/to_user */
#include "sba.h"



static int sba_major;

/*This holds the sba device related info*/
sba_dev sba_device;

/* Our request queue */
static struct request_queue *sba_queue;

static int dev;
// = MKDEV(8, 18);

static bio_end_io_t sba_end_io;

///////// map functions put here


/*The physical sector free list*/

char* bitmap; //[BITMAPSIZE];
struct timeval * prev_timeval;

int init_time(){
	prev_timeval = (struct timeval*) kmalloc(sizeof(struct timeval),GFP_KERNEL);
	if(!prev_timeval)
		return 1;
	else{
		prev_timeval->tv_sec = 0;
		prev_timeval->tv_usec =0;
	}
	return 0;
}


int init_bitmap(){
	int i;
	////////////////////////////////////
	//XXX: change this to kernel memory allocation routine
	bitmap = (char*) kmalloc(BITMAPSIZE,GFP_KERNEL);
	////////////////////////////////////
        if(!bitmap){
           printk(" Unable to allocate memory for bitmap\n");
           return 1;
        }
	for(i=0;i<BITMAPSIZE;i++)
		bitmap[i]=0;
	return 0;
}

/*Marks the corresponding physical sector as occupied*/

void set_occupied(sector_t sector){

	char field;
	int position = sector%8;
	sector_t index = (sector_t)(sector/8);
	char mask = 0x1;
	mask = mask << position;
	field = bitmap[index];
	bitmap[index] = field | mask;
	return;

}


/*Marks the corresponding physical sector as unoccupied*/

void set_unoccupied(sector_t sector){

	char field;
	int position = sector%8;
	sector_t index = (sector_t)(sector/8);
	char mask = 0x1;
	mask = mask << position;
	mask = ~mask;
	field = bitmap[index];
	bitmap[index] = field & mask;
	return;

}


/* returns zero if the sector is unoccupied and nonzero otherwise */

int is_occupied(sector_t sector){

	char field;
	int position = sector%8;
	sector_t index = (sector_t)(sector/8);
	char mask = 0x1;
	mask = mask << position;
	field = bitmap[index];
	return field & mask;

}


void release_bitmap(){
        if(bitmap)
		kfree(bitmap);
	return;
}

/*-----------------------------------------------*/

 

int init_l2p_map(){
	//int i;
	////////////////////////////////////////
	//XXX: change to actual memory handling routines of kernel
	l2p_map = (map_entry*) kmalloc(sizeof(map_entry),GFP_KERNEL);
	if(!l2p_map){
           printk(" Could not allocate memory for our l2p map\n");
           return 1;
        }
        l2p_map->l_sector = -1;
        l2p_map->p_sector = -1;
        l2p_map->version = -1;
        l2p_map->next = NULL;
	////////////////////////////////////////
	//for(i=0;i<SBA_SIZE;i++){
	//	l2p_map[i].phys_sect_no =-1;
	//	l2p_map[i].version = -1;
	//}
	return 0;
}

int set_map_entry(sector_t l_sector, sector_t p_sector){

        map_entry *newentry;
        map_entry *temp = l2p_map->next;
       	if(temp!=NULL){
		do{
          		if(temp->l_sector == l_sector){
             			temp->p_sector = p_sector;
             			temp->version ++;
             			return 0;
          		}
  	          	temp = temp->next;
        	}while(temp != NULL);
        }

	newentry = (map_entry*) kmalloc(sizeof(map_entry),GFP_KERNEL);
        if(!newentry){
	   printk(" Could not allocate memory for our l2p map\n");
           return 1;
        }
	newentry->p_sector = p_sector;
	newentry->l_sector = l_sector;
        newentry->version = 0;
        newentry->next = l2p_map->next;
        l2p_map->next = newentry;
	return 0;
}


map_entry* get_map_entry(sector_t l_sector){
	map_entry *temp = l2p_map->next;
        while(temp != NULL){
            if(temp->l_sector == l_sector)
               return temp;
            temp = temp->next;
        }
        return NULL;
}

void release_map(){
        if(l2p_map){
            map_entry *temp = l2p_map->next;
            while(temp!=NULL){
               		kfree(l2p_map);
               		l2p_map = temp;
               		temp = temp->next;
            	}
             kfree(l2p_map);
          }
	return;
}

void release_timer(){
	if(prev_timeval!= NULL)
		kfree(prev_timeval);
	return;
}


//////// map functions end here


sector_t newchunk(sector_t orig, int cnt, struct timeval * curr, struct timeval *  prev){
	//the algorithm to intelligently allocate physical blocks goes here

        int chunk_found = 0;
        int count = 0;


       while(chunk_found == 0){

		if(is_occupied(orig)){
			if(orig == JOURNAL_SIZE - 1)
				orig = 0;
			else
				orig ++;
			count ++;
                        
			if(count >= JOURNAL_SIZE)
	                      return -1; //// cant find

		}
                else {
	

		/*while((is_occupied(orig))&&(count<JOURNAL_SIZE)){
			if(orig==JOURNAL_SIZE-1)
				orig =0;
			else
				orig++;
			count++;
		}*/
        	//// here we get orig to be not occupied
                //// now we look for next "cnt-1" free sectors
               		int x=1;
       
        		while(x < cnt ){
				if(orig==JOURNAL_SIZE-1)
					orig =0;
				else
					orig++;
			
				if(is_occupied(orig))
					break;
				
				x++;
				count ++;
				if(count >= JOURNAL_SIZE)
					return -1; /// cant find
			
        		}
        		if(x != cnt){
            			orig = orig + 1;
				count = count + 1;

				if(count >= JOURNAL_SIZE)
					return -1;
		    	}
        		else
	  			chunk_found = 1;
			
		}
	}
        
	if(!is_occupied(orig))
    		return orig;
	else
		return -1; //XXX this should not happen, we need to chase this later.
}

/* Return time diff in micro seconds */
long long sba_common_diff_time(struct timeval st, struct timeval et)
{
        long long diff;
   
        if (et.tv_usec < st.tv_usec) {
                et.tv_usec += 1000000;
                et.tv_sec --;
        }

        diff = (et.tv_sec - st.tv_sec)*1000000 + (et.tv_usec - st.tv_usec);
       return diff;
}



/* Open the device. Increment the usage count */
int sba_open(struct inode *inode, struct file *filp)
{
        int num = iminor(inode);
        if (num >= SBA_DEVS) {
                return -ENODEV;
        }
        printk(" Device Opened YO \n");
        sba_device.usage ++;

        return 0;
}

/*Close the device. Decrement the usage count.*/
int sba_release(struct inode *inode, struct file *filp)
{
        struct block_device *bd = I_BDEV(inode);
        sba_device.usage --;

        printk(" Device Being Closed \n");
        if (!sba_device.usage) {
                fsync_bdev(bd);
        }

        return 0;
}

int sba_ioctl(struct inode *inode, struct file *filp,
                                unsigned int cmd, unsigned long arg)
{

	return -ENOTTY;
}

int sba_media_changed(struct gendisk *gd)
{
        return 1;
}

int sba_revalidate_disk(struct gendisk *gd)
{
        return 0;
}


struct block_device_operations sba_bdops = {
        owner: THIS_MODULE,
        open: sba_open,
        release: sba_release,
        ioctl: sba_ioctl,
        media_changed: sba_media_changed,
        revalidate_disk: sba_revalidate_disk
};

int sba_end_io(struct bio *sba_bio_clone, unsigned int bytes, int error)
{
	int uptodate;
	struct completion *event;
        uptodate = test_bit(BIO_UPTODATE, &sba_bio_clone->bi_flags);
	
	event = NULL;
	if (sba_bio_clone->bi_private) {
		// get sba_bio_orig from the private data structure .. 
		sba_req* sba_req_old = sba_bio_clone->bi_private;
                struct bio *sba_bio_orig;
		sba_bio_orig = sba_req_old->bio_old;
		event = sba_req_old->old_event;
                if(uptodate){
                        printk(" Update ");
                 	bio_endio(sba_bio_orig, sba_bio_orig->bi_size, 0);
			}
                 else{
                     printk(" Not uptodate ");
                     bio_endio(sba_bio_orig, sba_bio_orig->bi_size, -EIO);
		}
                 bio_put(sba_bio_clone);
                 
                 printk(" Ending up io \n");
		
		complete(event);//XXX: this frees the event ??
                 return 0;
        }
	else {
		// wouldn't generally expect this to happen
		//complete(event);//XXX: this frees the event ??
                bio_put(sba_bio_clone);
                return 1;
        }
       return 1;
}

int sba_print_bio(struct bio *sba_bio)
{
        int i;
       
        struct bio_vec *bvl;

        
        if (bio_data_dir(sba_bio) == READ) {
                printk( "READ block %ld size = %d sectors\n", (long int) SBA_SECTOR_TO_BLOCK(sba_bio->bi_sector), (int) bio_sectors(sba_bio));
		printk("vcnt: %d, idx: %d\n", sba_bio->bi_vcnt, sba_bio->bi_idx);
                //access each page of data
          bio_for_each_segment(bvl, sba_bio, i) {
                        printk( "READ: Page vir addrs %0x\n", (int)(bio_iovec_idx(sba_bio, i)->bv_page));
                }
        }
	else {
	
                printk( "WRITE block %ld size = %d sectors ", (long int) SBA_SECTOR_TO_BLOCK(sba_bio->bi_sector), (int) bio_sectors(sba_bio));
		//printk("vcnt: %d, idx: %d\n", sba_bio->bi_vcnt, sba_bio->bi_idx);
                
               // if (bio_sectors(sba_bio) > 8) { // why don't u have the if for READ ?
                        //access each page of data
                        bio_for_each_segment(bvl, sba_bio, i) {
                                printk( "WRITE: Page vir addrs %0x\n", (int)(bio_iovec_idx(sba_bio, i)->bv_page));
                        }
               // }
        }
       

        return 1;
}


int sba_new_request(request_queue_t *queue, struct bio *sba_bio)
{
	// create a data structure you want and store a pointer to it 
	// in sba_bio->bi_private ... you can access this again in the
	// end_io routine ....
	
	// clone the bio and get sba_bio_clone 
	// set a different end_io routine
	int index;
	int l_i;
        char *data = bio_data(sba_bio);
	//struct page *sba_page= NULL;		
        struct bio* sba_bio_clone;
	int num_sectors= bio_sectors(sba_bio);
	struct completion *event = NULL;
 	sba_req *sba_request;
	map_entry *me;
	struct timeval* curr_timeval = NULL;
	printk(" Request at sector %lu of size: %u  in sectors, and %u in bytes\n",(long int) /*SBA_SECTOR_TO_BLOCK*/(sba_bio->bi_sector), bio_sectors(sba_bio),sba_bio->bi_size);
        
	if(sba_bio->bi_next == NULL)
        	printk("XXXXX this is  a single block request \n");

         else
                printk(" will need to do the cloning change wahte \n"); 
        

        /// set the device
	sba_bio_clone = bio_clone(sba_bio,GFP_NOIO);
        sba_bio_clone->bi_bdev = sba_device.dev;
	////////////////////////////////////////
	/*Have to schedule a time event*/
	event = (struct completion*)kmalloc(sizeof(struct completion), GFP_KERNEL);
        if (!event) {
                printk("Error: mem allocation\n");
		//XXX: free all the memory
                return 1;
        }
        sba_request = kmalloc(sizeof(sba_request), GFP_KERNEL);
        if (!sba_request) {
               printk("Error: unable to allocate memory to sba request\n");
		//XXX: free all the memory
                return 1;
        }
	sba_request->bio_old = sba_bio;
	sba_request->old_event = event;
	sba_bio_clone->bi_private = sba_request;
	////////////////////////////////////////
	sba_bio_clone->bi_end_io = sba_end_io;
	//sba_print_bio(sba_bio_clone);
        sba_bio_clone->bi_rw = bio_data_dir(sba_bio);

        if(bio_data_dir(sba_bio) == READ){
        
        	printk("****** READ REQUEST****** \n"); // data is %s\n",data);
		 me=get_map_entry(sba_bio->bi_sector);//XXX: assume only one sector for now
		if(me){
			printk("@@@@@@ found matching logical entry @@@@@@@\n");
			sba_bio_clone->bi_sector = me->p_sector;
		}
		else{//XXX: this shouldn't happen ??
			printk("FATAL:: couldn't find matching logical entry\n");
			sba_bio_clone->bi_sector = sba_bio->bi_sector; 
		}
        }

        else{
        
        	printk("****** WRITE REQUEST****** \n data is %s\n",data);
		/**** get the time ***/
		curr_timeval = (struct timeval* )kmalloc(sizeof(struct timeval), GFP_KERNEL);
		if(!curr_timeval){
			printk("couldn't allocate memory for curr_timeval");
			//XXX: free all the memory
			return 1;
		}

		do_gettimeofday(curr_timeval);		
		//int write_size = 4096; //XXX: write size in bytes
		kfree(curr_timeval);
		
		 l_i = 0;
                while(l_i < num_sectors){
			map_entry *me = get_map_entry(sba_bio->bi_sector + l_i);
                	if(me){
                    		set_unoccupied(me->p_sector);
                	}
			l_i ++;
		}

		sba_bio_clone->bi_sector = newchunk(sba_bio->bi_sector,num_sectors,curr_timeval,prev_timeval); //// XXX : need to check if new sector fails, when there are no more sectors
		
		if(sba_bio_clone->bi_sector == -1){
			printk(" Error: could not find sector \n");
			return 1;
		}

		 index = 0;
		while(index<num_sectors){
                
			set_map_entry(sba_bio->bi_sector+index,sba_bio_clone->bi_sector+index);
			set_occupied(sba_bio_clone->bi_sector+index);
			index ++;
		}
        }
	printk("registering event\n");
       init_completion(event);


       	printk("generic make request\n");
        generic_make_request(sba_bio_clone);
	printk("wait for completion\n");
        wait_for_completion(event);
//////////////////////////////////////////////
	printk("got out\n");

 
//	generic_make_request(sba_bio_clone);

	return 0;

}
	
/* This method will initialize the device specific structures */
int __init sba_init(void)
{
        int nsectors;
	dev_t dev_no;
	int status;
	dev = MKDEV(8,5);
         //MKDEV(3,64);

	SBA_LOCK_INIT(&(sba_device.lock));
	sba_device.size = SBA_SIZE;
	dev_no = dev;
	sba_device.dev = open_by_devnum(dev_no, FMODE_READ|FMODE_WRITE);

	if (IS_ERR(sba_device.dev)) {
                goto out;
        }

	nsectors = sba_device.size * 2;
	
	/* Get a request queue */
        sba_queue = blk_alloc_queue(GFP_KERNEL);
        if (!sba_queue) {
                goto out;
        }
	blk_queue_hardsect_size(sba_queue, SBA_HARDSECT);
        blk_queue_make_request(sba_queue, sba_new_request);
	

        printk(" Major Number : %d\n",sba_major);
        /* Register */
        sba_major = register_blkdev(sba_major, DEVICE_NAME);
        if (sba_major <= 0) {
                printk( "sba: can't get major %d\n", sba_major);
                goto out;
        }

        printk( " got major number : %d\n",sba_major);

	/* Add the gendisk structure */
        sba_device.gd = alloc_disk(SBA_DEVS);
        if (!sba_device.gd) {
                goto out_unregister;
        }

	sba_device.gd->major = sba_major;
        sba_device.gd->first_minor = 0;
        sba_device.gd->fops = &sba_bdops;
        sba_device.gd->private_data = &sba_device;
        sba_device.gd->queue = sba_queue;
        strcpy(sba_device.gd->disk_name, DEVICE_NAME);
        set_capacity(sba_device.gd, nsectors);

	add_disk(sba_device.gd);

	/*initialize other data structures*/
	//timer
	status = init_time();
	if(status)
		goto out_unregister;
	
	//bitmap
        status = init_bitmap();
        if(status)
            goto out_unregister;
	
	//l2p map
        status = init_l2p_map();
        if(status)
            goto out_unmap;
 
	printk( "SBA init over ... successfully added the driver (total sec %d)\n", nsectors);
	return 0;

    out_unmap: release_bitmap();

    out_unregister:
        unregister_blkdev(sba_major, DEVICE_NAME);
    out:
        printk( "Unable to load the device\n");

        return -ENOMEM;
}

static __exit void sba_cleanup(void)
{
        del_gendisk(sba_device.gd);
        put_disk(sba_device.gd);
        unregister_blkdev(sba_major, DEVICE_NAME);
        blk_cleanup_queue(sba_queue);
        release_bitmap();
	release_timer();
        release_map();
        printk( "SBA cleanup over ... exiting\n");
}

module_init(sba_init);
module_exit(sba_cleanup);

MODULE_LICENSE("Dual BSD/GPL");


