
#ifndef __INCLUDE_SBA_H__
#define __INCLUDE_SBA_H__

//#include "sba_os_spec.h"

#define DEVICE_NAME     "sba"
#define SBA_MAJOR               0
#define SBA_DEVS                1       /*these are # devices exported by SBA*/

#define SBA_BLKSIZE		4096	    /* block size */
#define SBA_BLKSIZE_BITS        12          /* no of bits - BLKSIZE */
#define SBA_HARDSECT            512         /* sector size */
#define SBA_HARDSECT_BITS       9           /* no of bits - HARDSECT */


#define SBA_SECTOR_TO_BLOCK(a)  ((a) >> (SBA_BLKSIZE_BITS - SBA_HARDSECT_BITS))
//#define SBA_SIZE                1581232 /* no of 1 KB blocks in the sba device */
#define SBA_SIZE                104857 /* no of 1 KB blocks in the hdb device */
//#define SBA_SIZE                1606500 /* no of 1 KB blocks in the sba device */
#define SBA_LOCK_INIT(a)        spin_lock_init((a))
#define SBA_LOCK(a)                     spin_lock((a))
#define SBA_TRYLOCK(a)          spin_trylock((a))
#define SBA_UNLOCK(a)           spin_unlock((a))



#define JOURNAL_SIZE 104857 // number of sectors in physical media
				//XXX: have to remove this
#define BITMAPSIZE (int)(JOURNAL_SIZE/4)+1

typedef struct _sba_dev {
        struct block_device *dev;
        unsigned long size;
        spinlock_t lock;
        struct gendisk *gd;
        int usage;
}sba_dev;

typedef struct _sba_req {
	struct bio *bio_old;
	struct completion* old_event;
	
}sba_req;

typedef struct _map_entry{
        sector_t l_sector;
	sector_t p_sector;
	int version;
        struct _map_entry *next;
} map_entry;

map_entry *l2p_map;//[SBA_SIZE];



int init_time(void);
int init_bitmap(void);
void set_occupied(sector_t sector);
void set_unoccupied(sector_t sector);
int is_occupied(sector_t sector);
void release_bitmap(void);
int init_l2p_map(void);
int set_map_entry(sector_t l_sector, sector_t p_sector);
map_entry* get_map_entry(sector_t l_sector);
void release_map(void);
void release_timer(void);
sector_t newchunk(sector_t orig, int cnt, struct timeval * curr, struct timeval * prev);
long long sba_common_diff_time(struct timeval st, struct timeval et);

int sba_open(struct inode *inode, struct file *filp);
int sba_release(struct inode *inode, struct file *filp);
int sba_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg);
int sba_media_changed(struct gendisk *gd);
int sba_revalidate_disk(struct gendisk *gd);

#endif /* __INCLUDE_SBA_H__ */

