#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/uio.h>
#include <assert.h>

#define _GNU_SOURCE
#include <getopt.h>

//#include "ioctl.h"
#include <sys/ioctl.h>


#define BLKRASET   _IO(0x12,98) /* set read ahead for block device */
#define BLKRAGET   _IO(0x12,99) /* get current read ahead setting */
#define BLKFRASET  _IO(0x12,100)/* set filesystem (mm/filemap.c) read-ahead */
#define BLKFRAGET  _IO(0x12,101)/* get filesystem (mm/filemap.c) read-ahead */

// max 1 MB
#define MAX_BUFFER 1048576
#define PAGE_SIZE 4096
#define MIN_OFFSET 0
/* 1.5 GB */
#define MAX_OFFSET 1610612736 ///// should be the size of disk.
#define MAX_SAMPLES 100000
#define MAX_BUF 104857600
#define MAX_OFFSET_LOG (52428800) /// this is the maximum offset with respect to log  where u can write 
//#define SKIP_LOG (52428800 * 20)
#define SKIP_LOG 0
//#define SKIP_LOG (1000 * 1024 * 1024)

double start_bench, end_bench;
int fd_raw;
char * buffer;

double avg_service_time = 0;
double min_service_time = 0;
double max_service_time = 0;


double bench_write_time[MAX_SAMPLES];
double bench_sync_time[MAX_SAMPLES];
char buf[MAX_BUF];


int size_write = -1;
int size_skip = -1;
int us_sleep = -1;
int repeat = 0;
int log_offset = 0;

double total_data_written = 0;
double total_write_time = 0;
double total_bench_time = 0;


void clear_cache();


#ifdef RDTSC
/* Time to taken by a tic on the db cluster 551.261 MHz */
#define MICROSEC_PER_TIC 0.00181488
#define GET_TIME(t) asm("rdtsc" : "=A" (t)); t = t * MICROSEC_PER_TIC
#define GET_MSEC_TIME(t) asm("rdtsc" : "=A" (t)); t = t * MICROSEC_PER_TIC / 1000
#else
#define GET_TIME(t) t = get_microsec()
#define GET_MSEC_TIME(t) t = get_microsec(); t = t / 1000
#endif



double get_microsec() {
  struct timeval timeVal;
  double micro = -1;

  gettimeofday(&timeVal, 0);
  micro = (double)timeVal.tv_sec * 1000000 + timeVal.tv_usec;

  return micro;
}



void init(char * raw_device, int random_seed){
  int addr;

  //if ((fd_raw = open(raw_device, O_RDWR|O_SYNC)) < 0){
  if ((fd_raw = open(raw_device, O_RDWR)) < 0){
    perror("Error opening input raw device");
    fprintf(stderr, "%s\n", raw_device);
    exit(-1);
  }
  
  buffer = malloc(MAX_BUFFER + 8192);
  addr = (unsigned int) buffer;
  buffer = (unsigned char *)(addr + PAGE_SIZE - addr % (PAGE_SIZE));
  srand(random_seed);

}




void clear_cache(){
  int i;
  double temp;
  int offset;
  int res;
  

  for (i = 0; i < 10; i++){
    temp = (double)rand()/(RAND_MAX + 1.0);
    offset = MIN_OFFSET + (MAX_OFFSET - MIN_OFFSET) * temp;
    offset = ( offset / PAGE_SIZE ) * PAGE_SIZE;
    res = pread(fd_raw, buffer, 4096, offset);
    
    if ( res < 0 ){
      perror("Error reading/writing from file");
      fprintf(stderr, "Error reading %d current offset, size %d\n", offset, 4096);
      fflush(stderr);
      exit(-1);
    }
  }

}



/* for repeat steps:
   read size_chunk data (KB), then skip size_gap (KB) and sleep for us_sleep 
   
 */
int quasi_seq_writes(int size_write, int size_skip, int us_sleep, int repeat, int log_offset){
  int i = 0;
  double start_time_write, end_time_write, end_time_sync;
  int res = 0;
  double start_bench, end_bench; 
  double start_time_sleep, end_time_sleep;
  double wait_time;
  int j = 0;
  unsigned int crt_offset = 0;


  //crt_offset = log_offset * 1024;
  lseek(fd_raw, crt_offset + SKIP_LOG + log_offset * 1024, SEEK_SET);

  /* add timing for all these operations */
  GET_TIME(start_bench);
  for (i = 0; i < repeat; i++){

    GET_TIME(start_time_write);    
    res = write(fd_raw, buffer, size_write * 1024);
    if ((res == 0) || (res == -1)) {
      printf("Error writing data %d\n", res);
      exit(-1);
    }	
    GET_TIME(end_time_write);
    bench_write_time[i] = end_time_write - start_time_write;
    fsync(fd_raw);
    GET_TIME(end_time_sync);
    bench_sync_time[i] = end_time_sync - end_time_write;
    crt_offset = crt_offset + size_write * 1024 + size_skip * 1024;
    if ((crt_offset - size_write * 1024) >= MAX_OFFSET_LOG) {
      crt_offset = crt_offset % MAX_OFFSET_LOG;
    }
    lseek(fd_raw, crt_offset + SKIP_LOG + log_offset * 1024, SEEK_SET);
    //wait for us_sleep us
    GET_TIME(start_time_sleep);
    wait_time = 0;
    while (wait_time < us_sleep) {
      //do nothing
      j = j + 1;
      GET_TIME(end_time_sleep);
      wait_time = end_time_sleep - start_time_sleep;
    }

    //custom_sleep(us_sleep);
    total_data_written += ((float)size_write) / 1024; // in MB
    total_write_time += bench_write_time[i] / 1000000; // in seconds
  }
  GET_TIME(end_bench);
  total_bench_time = (end_bench - start_bench) / 1000000;
  return j;
}


/* for repeat steps:
   read size_chunk data (KB), then skip size_gap (KB) and sleep for us_sleep 
   
 */
int rand_positioning_quasi_seq_writes(int size_write, int size_skip, int us_sleep, int repeat, int log_offset){
  int i = 0;
  double start_time_write, end_time_write, end_time_sync;
  int res = 0;
  double start_bench, end_bench; 
  double start_time_sleep, end_time_sleep;
  double wait_time;
  int j = 0;
  unsigned int crt_offset = 0;
  unsigned int random_offset = 0;

  
  srand(0);
  
  //crt_offset = log_offset * 1024;
  lseek(fd_raw, crt_offset + SKIP_LOG + log_offset * 1024, SEEK_SET);

  /* add timing for all these operations */
  GET_TIME(start_bench);
  for (i = 0; i < repeat; i++){

    /* position to a random location in the log and perform a write */
    random_offset = (unsigned int) (
      ((double)(MAX_OFFSET_LOG - 2 * size_write * 1024 - size_skip * 1024) * 
       (rand()/(RAND_MAX + 1.0))) 
      + SKIP_LOG + log_offset * 1024);
    random_offset = (random_offset / 4096) * 4096;
    lseek(fd_raw, random_offset, SEEK_SET);

    //printf("Write to random offset %d size %d\n", random_offset / 1024, size_write);

    res = write(fd_raw, buffer, size_write * 1024);
    if ((res == 0) || (res == -1)) {
      printf("Error writing data %d\n", res);
      exit(-1);
    }

    //lseek(fd_raw, crt_offset + SKIP_LOG + log_offset * 1024, SEEK_SET);
    //wait for us_sleep us
    GET_TIME(start_time_sleep);
    wait_time = 0;
    while (wait_time < us_sleep) {
      //do nothing
      j = j + 1;
      GET_TIME(end_time_sleep);
      wait_time = end_time_sleep - start_time_sleep;
    }

    /* skip size_skip KB and then perform a write that we are going to time */
    lseek(fd_raw, random_offset + size_skip * 1024 + size_write * 1024, SEEK_SET);

    //printf("Write to %d after skipping %d size %d\n", (random_offset + size_skip * 1024 + size_write * 1024) / 1024, size_skip, size_write);

    GET_TIME(start_time_write);    
    res = write(fd_raw, buffer, size_write * 1024);
    if ((res == 0) || (res == -1)) {
      printf("Error writing data %d\n", res);
      exit(-1);
    }	
    GET_TIME(end_time_write);
    bench_write_time[i] = end_time_write - start_time_write;
    fsync(fd_raw);
    GET_TIME(end_time_sync);
    bench_sync_time[i] = end_time_sync - end_time_write;
    /*
    crt_offset = crt_offset + size_write * 1024 + size_skip * 1024;
    if ((crt_offset - size_write * 1024) >= MAX_OFFSET_LOG) {
      crt_offset = crt_offset % MAX_OFFSET_LOG;
    }
    */

    //custom_sleep(us_sleep);
    total_data_written += ((float)size_write) / 1024; // in MB
    total_write_time += bench_write_time[i] / 1000000; // in seconds
  }
  GET_TIME(end_bench);
  total_bench_time = (end_bench - start_bench) / 1000000;
  return j;
}



void compute_stats(int repeat) {
  int i;
  double sum = 0;

  min_service_time = 100000;
  max_service_time = 0; 

  for (i = 0; i < repeat; i++){
    sum = sum + bench_write_time[i] / 1000;
    if (min_service_time > bench_write_time[i])
      min_service_time = bench_write_time[i];
    if (max_service_time < bench_write_time[i])
      max_service_time = bench_write_time[i];
  }
  avg_service_time = sum / repeat;
  min_service_time = min_service_time / 1000;
  max_service_time = max_service_time / 1000;

}



void print_stats(int size_write, int repeat){
  int i = 0;

  printf("Times for writing data\n");
  for (i = 0; i < repeat; i++){
    printf("Time to write %d KB: %f ms\n", size_write, bench_write_time[i] / 1000);
  }
  /*
  printf("Times for synching data:");
  for (i = 0; i < repeat; i++){
    printf("Time to sync %d KB: %f ms\n", size_write, bench_sync_time[i] / 1000);
  }
  */

}


void process_arg(const char * option, char * value){
  //  printf("option %s value %s\n", option, value);

  if (!strcmp(option, "size_write")){
    size_write = atoi(value);
    if (size_write <= 0){
      printf("Error,  write size too small %d\n", size_write);
      fflush(stdout);
      exit(-1);
    }
    return;
  }

  if (!strcmp(option, "log_offset")){ /* in KB */
    log_offset = atoi(value);
    if (log_offset < 0){
      printf("Error,  log_offset too small %d\n", log_offset);
      fflush(stdout);
      exit(-1);
    }
    return;
  }

  if (!strcmp(option, "size_skip")){
    size_skip = atoi(value);
    if (size_skip < 0){
      printf("Error,  skip size too small %d\n", size_skip);
      fflush(stdout);
      exit(-1);
    }
    return;
  }

  if (!strcmp(option, "us_sleep")){
    us_sleep = atoi(value);
    if (us_sleep < 0){
      printf("Error,  sleep time too small %d\n", us_sleep);
      fflush(stdout);
      exit(-1);
    }
    return;
  }

  if (!strcmp(option, "repeat")){
    repeat = atoi(value);
    if (repeat <= 0){
      printf("Error,  repeat too small %d\n", repeat);
      fflush(stdout);
      exit(-1);
    }
    return;
  }
}



struct option long_options[] = {  
  {"size_write", 1, 0, 0}, 
  {"size_skip", 1, 0, 0},
  {"us_sleep", 1, 0, 0},
  {"repeat", 1, 0, 0},
  {"log_offset", 1, 0, 0}
};




int main(int argc, char ** argv){  
  unsigned int val;
  int c;
  int option_index = 0;
  int fd;
  
  /*
  if ((fd = open("/dev/sdd", O_RDWR)) < 0){
    perror("Error opening input raw device");
    fprintf(stderr, "%s\n", fd);
    exit(-1);
  }
  
  ioctl(fd, BLKRAGET, &val);
  printf("val %u\n", val);
  exit(0);
  */
  while (1){
    c = getopt_long (argc, argv, "",
		     long_options, &option_index);
				     
    if (c == -1)
      break;

	switch (c){
	  case 0:
	    process_arg(long_options[option_index].name, optarg);
																				        break;            
    }    
  }

  assert( (size_write > 0) && (size_skip >= 0) && (us_sleep >= 0) && (repeat >= 0));

  init("/dev/hdb", 0);
  clear_cache();
  /* 1 k writes, no gap, no sleep */
  rand_positioning_quasi_seq_writes(size_write, size_skip, us_sleep, repeat, log_offset);
  //print_stats(size_write, repeat);
  compute_stats(repeat);
  printf("Results indiv_write %d KB skip %d KB ms_sleep %f ms repeat %d total_data_written %f MB total_write_time %f s total_bench_time %f s bandwidth %f MB/s log_offset %d avg_service_time %f ms min_service_time %f ms max_service_time %f ms\n", size_write, size_skip, ((float)(us_sleep))/1000, repeat, total_data_written, total_write_time, total_bench_time, total_data_written/total_write_time, log_offset, avg_service_time, min_service_time, max_service_time);
  
  return 0;
}














